@@ -8,41 +8,50 @@ import time
 import warnings
 from collections import OrderedDict
 from datetime import datetime
-from urllib.error import HTTPError
+from typing import Dict
 from urllib.parse import quote
-from urllib.request import urlopen
 
 import numpy as np
+import pandas as pd
 import pubchempy as pcp
+import quantities as pq
+import requests
 from IPython.display import display
 from PIL import Image
-
-import quantities as pq
-from pyrfume import load_data, logger, tqdm, trange
-from pyrfume.physics import mackay
 from quantities.constants.statisticalmechanics import R
+from tqdm.auto import tqdm, trange
+
+from pyrfume import load_data, logger
+from pyrfume.physics import mackay
 
 try:
-    from rdkit import Chem
-    from rdkit.Chem import Draw, AllChem, SaltRemover
+    from rdkit import Chem, RDLogger
+    from rdkit.Chem import AllChem, Draw, SaltRemover
+
+    rdkit_logger = RDLogger.logger()
+    RDKIT = True
 except ImportError:
     warnings.warn(
         "Parts of mordred and/or rdkit could not be imported; try installing rdkit via conda",
         UserWarning,
     )
+    RDKIT = False
 
 ROOM_TEMP = (22 + 273.15) * pq.Kelvin
 ROOM_PRESSURE = 1 * pq.atm
 GAS_MOLAR_DENSITY = ROOM_PRESSURE / (R * ROOM_TEMP)
 
-ODORANTS_BASIC_INFO_PATH = "odorants/all-cids-properties.csv"
-ODORANT_SOURCES_PATH = "odorants/all-cids.csv"
+ODORANTS_BASIC_INFO_PATH = "molecules/molecules.csv"
+ODORANT_SOURCES_PATH = "molecules/usage.csv"
+
+PUBCHEM_KINDS = ["name", "smiles", "inchi", "inchikey", "formula", "sdf", None]
 
 
 class Solution:
-    components: dict = {}
-    
-    def __init__(self, components: dict, date_created: datetime.timestamp = None):
+    components: Dict["Compound", pq.quantity.Quantity] = None
+    date_created: datetime = None
+
+    def __init__(self, components: dict, date_created: datetime = None):
         self.total_volume = 0 * pq.mL
         assert isinstance(components, dict), "Components must be a dict"
         for component, volume in components.items():
@@ -55,6 +64,8 @@ class Solution:
                 raise ValueError("Components must be provided with volumes")
             self.total_volume += volume  # Assume that volume is conserved
         self.components = components
+        if not date_created:
+            date_created = str(datetime.now())[:-7]
         self.date_created = date_created if date_created else datetime.now()
 
     @property
@@ -131,7 +142,9 @@ class Solution:
     def total_pressure(self):
         """Computes total pressure of the vapor using Dalton's law"""
         preferred_units = pq.Pa
-        partial_pressures = [pressure.rescale(preferred_units) for pressure in self.partial_pressures.values()]
+        partial_pressures = [
+            pressure.rescale(preferred_units) for pressure in self.partial_pressures.values()
+        ]
         return preferred_units * np.sum(partial_pressures)
 
     @property
@@ -163,10 +176,11 @@ class Solution:
 
 
 class Molecule:
-    def __init__(self, cid: int, name: str=None, fill: bool=False):
+    def __init__(self, cid: int, name: str = None, fill: bool = False, delay=0):
         self.cid = cid
+
         if fill:
-            self.fill_details()
+            self.fill_details(delay=delay)
         if name:
             self.name = name
 
@@ -200,14 +214,14 @@ class Molecule:
     def molar_evaporation_rate(self):
         return mackay(self.vapor_pressure)
 
-    def fill_details(self):
+    def fill_details(self, delay=0):
         assert self.cid is not None
         url_template = (
             "https://pubchem.ncbi.nlm.nih.gov/" "rest/pug/compound/cid/%d/property/" "%s/JSON"
         )
         property_list = ["MolecularWeight", "IsomericSMILES"]
         url = url_template % (self.cid, ",".join(property_list))
-        json_data = url_to_json(url)
+        json_data = url_to_json(url, delay=delay)
         details = json_data["PropertyTable"]["Properties"][0]
 
         def convert(name):
@@ -219,7 +233,7 @@ class Molecule:
                 assert value == self.cid, "REST API CID does not match provided CID"
             key = convert(key)
             if key == "molecular_weight":
-                value *= pq.g / pq.mol
+                value = float(value) * pq.g / pq.mol
             setattr(self, key, value)
 
         if not self.name:
@@ -284,7 +298,14 @@ class Vendor:
 
 
 class ChemicalOrder:
-    def __init__(self, molecule: "Molecule", vendor: "Vendor", part_id: str, purity: float=1, known_impurities: list=None):
+    def __init__(
+        self,
+        molecule: "Molecule",
+        vendor: "Vendor",
+        part_id: str,
+        purity: float = 1,
+        known_impurities: list = None,
+    ):
         self.molecule = molecule
         self.vendor = vendor
         self.part_id = part_id
@@ -305,8 +326,14 @@ class ChemicalOrder:
 
 class Compound:
     def __init__(
-        self, chemical_order: ChemicalOrder, stock: str="", date_arrived: datetime=None, date_opened: datetime=None, is_solvent: bool=False
+        self,
+        chemical_order: ChemicalOrder,
+        stock: str = "",
+        date_arrived: datetime = None,
+        date_opened: datetime = None,
+        is_solvent: bool = False,
     ):
+
         self.chemical_order = chemical_order
         self.stock = stock
         self.date_arrived = date_arrived if date_arrived else datetime.now
@@ -333,27 +360,53 @@ class Compound:
             return getattr(self.chemical_order.molecule, attr)
 
 
-def url_to_json(url, verbose=True) -> str:
+def url_to_json(url, verbose=True, delay=0) -> str:
     json_data = None
-    msgs = []
-    try:
-        page = urlopen(url)
-        string = page.read().decode("utf-8")
+    time.sleep(delay)
+    response = requests.get(url)
+    if response.status_code == 200:
+        string = response.content.decode("utf-8")
         json_data = json.loads(string)
-    except HTTPError:
-        msgs.append("HTTPError for query '%s'" % url)
-    if json_data is None and verbose:
-        for msg in msgs:
-            print(msg)
+    else:
+        msg = "HTTP Status Code %d for %s" % (response.status_code, url)
+        if verbose:
+            logger.error(msg)
     return json_data
 
 
+def is_kind(identifier: str, kind: str) -> bool:
+    if kind == "smiles":
+        if RDKIT:
+            rdkit_logger.setLevel(RDLogger.CRITICAL)
+            result = Chem.MolFromSmiles(identifier) is not None
+            rdkit_logger.setLevel(RDLogger.WARNING)
+        else:
+            result = None
+    elif kind == "inchikey":
+        result = len(identifier) == 27 and identifier[14] == "-" and identifier[25] == "-"
+    elif kind == "inchi":
+        if RDKIT:
+            result = Chem.inchi.MolFromInchi(identifier, logLevel=None) is not None
+        else:
+            result = None
+    elif kind == "name":
+        result = True
+    else:
+        result = False
+    return result
+
+
+def get_kind(identifier: str):
+    kinds = [kind for kind in PUBCHEM_KINDS if is_kind(identifier, kind)]
+    return kinds[-1]  # Return most sophisticated kind ('name' will always be in the list)
+
+
 def deisomerize_smiles(smiles: str) -> str:
     mol = Chem.MolFromSmiles(smiles)
     if mol:  # If a mol object was successfully create (i.e. not `None`)
         smiles = Chem.MolToSmiles(mol, isomericSmiles=False)
     else:
-        smiles = smiles.replace('@','').replace('@@','').replace('/','').replace('\\','')
+        smiles = smiles.replace("@", "").replace("@@", "").replace("/", "").replace("\\", "")
     return smiles
 
 
@@ -371,18 +424,23 @@ def canonical_smiles(smiles: str, kekulize: bool = False) -> str:
 
 def get_cids(
     identifiers: list,
-    kind: str = "name",
+    kind: str = None,
     verbose: bool = True,
     wait: float = 0,
     results: dict = None,
 ) -> dict:
     """Return CIDs for molecule based on any synonym,
     including a chemical name or a CAS"""
-    kind = kind.lower()
+    if isinstance(kind, str):
+        kind = kind.lower()
+    assert kind in PUBCHEM_KINDS
     if results is None:
         results = {}
     p = tqdm(identifiers)
     for identifier in p:
+        # if not isinstance(identifier, str):
+        #    logger.warning("%s is not a string" % identifier)
+        #    continue
         p.set_description(str(identifier))
         cid = get_cid(identifier, kind=kind, verbose=verbose)
         if not cid:
@@ -393,19 +451,39 @@ def get_cids(
     return results
 
 
-def get_cid(
-    identifier: str, kind: str = "name", verbose: bool = True, fix_smiles_on_error: bool = True
+def get_cid(  # noqa: C901 (too complex -- TODO)
+    identifier: str,
+    kind: str = None,
+    verbose: bool = True,
+    fix_smiles_on_error: bool = True,
+    attempt=0,
 ) -> int:
-    """Return data about a molecule from any synonym,
-    including a chemical name or a CAS"""
-    kind = kind.lower()
+    """
+    Return data about a molecule from any synonym,
+    including a chemical name or a CAS.
+    """
     if isinstance(identifier, float) and np.isnan(identifier):
         return 0
+    replace = [("α", "alpha"), ("β", "beta"), ("γ", "gamma"), ("δ", "delta")]
+    for a, b in replace:
+        identifier = identifier.replace(a, b)
+    if kind is None:
+        kind = get_kind(identifier)
+    else:
+        kind = kind.lower()
     try:
         result = pcp.get_cids(identifier, namespace=kind)
     except pcp.BadRequestError:
         logger.warning('Request Error for "%s"' % identifier)
         result = []
+    except pcp.PubChemHTTPError as e:
+        if attempt == 0:
+            import time
+
+            time.sleep(10)
+            return get_cid(identifier, kind, verbose, fix_smiles_on_error, 1)
+        else:
+            raise e
     if not len(result):
         cid = 0
     else:
@@ -427,7 +505,7 @@ def from_cids(cids: list, property_list: bool = None) -> list:
     chunk_size = 100
     for start in trange(0, len(cids), chunk_size):
         stop = min(start + chunk_size, len(cids))
-        print("Retrieving %d through %d" % (start, stop - 1))
+        logger.info("Retrieving %d through %d" % (start, stop - 1))
         cid_subset = [
             str(x) for x in [cid for cid in cids[start:stop] if int(cid) > 0 and cid is not None]
         ]
@@ -444,11 +522,14 @@ def from_cids(cids: list, property_list: bool = None) -> list:
         )
         url = synonyms_template % (cid_subset)
         json_data = url_to_json(url)
-        information = json_data["InformationList"]["Information"]
+        if json_data is not None:
+            information = json_data["InformationList"]["Information"]
+        else:
+            information = []
         for i, d in enumerate(data):
             try:
                 synonyms = information[i]["Synonym"]
-            except KeyError:
+            except (IndexError, KeyError):
                 try:
                     d["name"] = d["IUPACName"]
                 except KeyError:
@@ -502,20 +583,28 @@ def cas_from_synonyms(synonyms: list) -> list:
 
 def cactus(identifier: str, output: str = "cas") -> str:
     url_template = "https://cactus.nci.nih.gov/chemical/structure/%s/%s"
-    identifier = identifier.replace(' ', '%20')
+    identifier = identifier.replace(" ", "%20")
     url = url_template % (identifier, output)
-    page = urlopen(url)
-    result = page.read().decode("utf-8")
+    response = requests.get(url)
+    if response.status_code == 200:
+        result = response.content.decode("utf-8")
+    else:
+        logger.error("HTTP Status Code %d for %s" % (response.status_code, url))
+        result = None
     return result
 
 
 def cactus_image(smiles: str) -> None:
     url_template = "https://cactus.nci.nih.gov/chemical/structure/%s/image"
-    smiles = smiles.replace(' ', '%20')
+    smiles = smiles.replace(" ", "%20")
     url = url_template % smiles
-    image_data = urlopen(url).read()
-    image = Image(image_data)
-    display(image)
+    response = requests.get(url)
+    if response.status_code == 200:
+        image_data = response.content.decode("utf-8")
+        image = Image(image_data)
+        display(image)
+    else:
+        logger.error("HTTP Status Code %d for %s" % (response.status_code, url))
 
 
 def get_compound_summary(cid: int, heading: str):
@@ -575,6 +664,38 @@ def _parse_other_info(info, records=None):
     return records
 
 
+def display_molecules(molecules: pd.DataFrame, no_of_columns=5, figsize=(15, 15)):
+    import matplotlib.pyplot as plt
+
+    fig = plt.figure(figsize=figsize)
+    column = 0
+    for i, (cid, info) in enumerate(molecules.iterrows()):
+        column += 1
+        #  check for end of column and create a new figure
+        if column == no_of_columns + 1:
+            fig = plt.figure(figsize=figsize)
+            column = 1
+        fig.add_subplot(1, no_of_columns, column)
+        image = smiles_to_image(info["IsomericSMILES"], png=False)
+        plt.imshow(image)
+        plt.axis("off")
+        plt.title("%d: %s" % (cid, info["name"]))
+
+
+def embed_molecules(molecules: pd.DataFrame):
+    import matplotlib.pyplot as plt
+
+    plt.figure(figsize=(6, 6))
+    ax = plt.gca()
+    embedding = load_data("embedding/pf_umap.pkl")
+    ax = embedding.plot.scatter(x=0, y=1, alpha=0.05, c="k", ax=ax)
+    smiles = molecules["IsomericSMILES"]
+    embedding_ = embedding.loc[smiles]
+    embedding_.plot.scatter(x=0, y=1, alpha=1, c="r", s=100, ax=ax)
+    ax.set_xlabel("Dimension 1")
+    ax.set_ylabel("Dimension 2")
+
+
 def smiles_to_image(smiles, png=True, b64=False, crop=True, padding=10, size=300):
     """
     png: Whether to convert to .png data (or to leave as a PIL image)
@@ -594,11 +715,13 @@ def smiles_to_image(smiles, png=True, b64=False, crop=True, padding=10, size=300
     return image
 
 
-def smiles_to_mol(smiles: list, max_attempts: int=10, use_random_coords: bool=False, deisomerize=False) -> dict:
+def smiles_to_mol(
+    smiles: list, max_attempts: int = 10, use_random_coords: bool = False, deisomerize=False
+) -> dict:
     if deisomerize:
         f = deisomerize_smiles
     else:
-        f = lambda x: x
+        f = lambda x: x  # noqa: E731 (don't assign lambda -- TODO)
     mols_raw = [Chem.MolFromSmiles(f(smi)) for smi in smiles]
     logger.info("Computing 3D coordinates...")
     s = SaltRemover.SaltRemover()
@@ -661,7 +784,7 @@ def all_smiles():
     """All SMILES found in the file at ODORANTS_BASIC_INFO_PATH.
     May contain duplicates (if two CIDs give the same SMILES)"""
     df = all_odorants()
-    return list(df["SMILES"])
+    return list(df["IsomericSMILES"])
 
 
 if __name__ == "__main__":
